"""Module for checking file status.
"""

import time
import os
import logging

from ox_mon.common import exceptions, configs, interface


class UnexpectedFileStatus(exceptions.OxMonAlarm):
    """Exception to indicate something was incorrect with file status.
    """


class SimpleFileChecker(interface.Checker):
    """Checker for file status (existence, non-existence, age, etc.)
    """

    @classmethod
    def options(cls):
        logging.debug('Making options for %s', cls)
        result = configs.BASIC_OPTIONS + [
            configs.OxMonOption(
                'target', help=('Path to target file to check.')),
            configs.OxMonOption(
                '--live', default=False, type=bool, is_flag=1, help=(
                    'If this option is given then file must exist.')),
            configs.OxMonOption(
                '--dead', default=False, type=bool, is_flag=1, help=(
                    'If this option is given then file must not exist.')),
            configs.OxMonOption(
                '--max-age-in-hours', type=float, help=(
                    'If given and file exists and is older then alarm raised.'
                    )),
            configs.OxMonOption(
                '--max-age-in-days', type=float, help=(
                    'If given and file exists and is older then alarm raised.'
                    ))
            ]
        return result

    def _check_liveness(self):
        """Check if existence correct.
        """
        if self.config.live:
            if not os.path.exists(self.config.target):
                msg = 'Required file does not exist: %s' % self.config.target
                raise UnexpectedFileStatus(msg)
        if self.config.dead:
            if os.path.exists(self.config.target):
                msg = 'Forbidden file present: %s' % self.config.target
                raise UnexpectedFileStatus(msg)

    def _check_age(self):
        """Check if file age is reasonable.
        """
        if not os.path.exists(self.config.target):
            return  # File does not exist so cannot check max age items
        stat_info = None
        if self.config.max_age_in_hours is not None:
            stat_info = os.stat(self.config.target)
            self._check_max_age_hours(stat_info)
        if self.config.max_age_in_days is not None:
            stat_info = stat_info if stat_info else os.stat(self.config.target)
            self._check_max_age_days(stat_info)

    def _check_max_age_hours(self, stat_info):
        """Take result of scan system call and check age in hours.

Should be called by _check method an dnot directly.
        """
        age_in_seconds = time.time() - stat_info.st_mtime
        age_in_hours = age_in_seconds / 3600.0
        if age_in_hours > self.config.max_age_in_hours:
            msg = 'File age %.2f hours > %.2f hours: %s' % (
                age_in_hours, self.config.max_age_in_hours, self.config.target)
            raise UnexpectedFileStatus(msg)

    def _check_max_age_days(self, stat_info):
        """Take result of scan system call and check age in days.

Should be called by _check method an dnot directly.
        """
        age_in_seconds = time.time() - stat_info.st_mtime
        age_in_days = age_in_seconds / 86400.0
        if age_in_days > self.config.max_age_in_days:
            msg = 'File age %.2f days > %.2f days: %s' % (
                age_in_days, self.config.max_age_in_days, self.config.target)
            raise UnexpectedFileStatus(msg)

    def _check(self):
        """Check file liveness, age, etc.
        """
        self._check_liveness()
        self._check_age()

        return 'Status as expected.'



"""Module for checking versions of things.
"""

import re
import subprocess
import logging

from ox_mon.common import exceptions, configs, interface


class UnexpectedVersion(exceptions.OxMonAlarm):
    """Exception to indicate version was wrong.
    """


class SimpleVersionChecker(interface.Checker):
    """Checker for versions of things.
    """

    @classmethod
    def options(cls):
        logging.debug('Making options for %s', cls)
        result = configs.BASIC_OPTIONS + [
            configs.OxMonOption(
                '--cmd', help=('Command to run to generate version')),
            configs.OxMonOption(
                '--flags', default='::version', help=(
                    'Comma separated list of options to pass to --cmd. '
                    'We replace colons with dashes so e.g., :v becomes -v.')),
            configs.OxMonOption(
                '--vre', default='[0-9]+[.][0-9]+[.]?[0-9]*', help=(
                    'Regular expression to find version in --cmd output.'
                    )),
            configs.OxMonOption('--exact', default=None, help=(
                'If provided, version must exactly match this.')),
            configs.OxMonOption('--cmpmode', default='int', help=(
                'If "str" compare as strings. If int compare as integers. '
                'In either case, first we split on the "." character.')),
            configs.OxMonOption('--minv', default=None, help=(
                'If given then this is minimum allowed version with '
                'comparison done based on --cmpmode.')),
            configs.OxMonOption('--maxv', default=None, help=(
                'If given then this is maximum allowed version with '
                'comparison done based on --cmpmode.')),
            ]
        return result

    def _get_version(self, cmd_out: str):
        """Find and return version string from command output.
        """
        my_re = re.compile(self.config.vre)
        match = my_re.search(cmd_out)
        return match

    def _compare_version(self, version):
        if self.config.exact is not None:
            if self.config.exact != version:
                raise UnexpectedVersion(
                    'Expected version "%s" but got "%s"' % (
                        self.config.exact, version))

        if self.config.minv:
            split_version = version.split('.')
            minv = self.config.minv.split('.')
            if self.config.cmpmode == 'int':
                minv = list(map(int, minv))
                split_version = list(map(int, split_version))
            elif self.config.cmpmode != 'str':
                raise ValueError('Invalid value for cmpmode: "%s"' % (
                    self.config.cmpmode))
            if split_version < minv:
                raise UnexpectedVersion('Version "%s" < minimum "%s"' % (
                    version, self.config.minv))
        if self.config.maxv:
            split_version = version.split('.')
            maxv = self.config.maxv.split('.')
            if self.config.cmpmode == 'int':
                maxv = list(map(int, maxv))
                split_version = list(map(int, split_version))
            elif self.config.cmpmode != 'str':
                raise ValueError('Invalid value for cmpmode: "%s"' % (
                    self.config.cmpmode))
            if split_version > maxv:
                raise UnexpectedVersion('Version "%s" > maximum "%s"' % (
                    version, self.config.maxv))

    def _check(self):
        """Check file liveness, age, etc.
        """
        cmd = [self.config.cmd]
        for item in self.config.flags.split(','):
            cmd.append(item.replace(':', '-'))
        proc = subprocess.run(cmd, check=False, stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE)
        if proc.returncode != 0:
            stdout = proc.stdout.decode('utf8')
            stderr = proc.stderr.decode('utf8')

            raise ValueError('Bad return code %s from %s:\n%s' % (
                proc.returncode, cmd, (
                    stdout if stdout else '') + '\n' + (
                        stderr if stderr else '')))

        # Need to use getattr to avoid strange pytype errors in lines below
        cmd_out = getattr(proc, 'stdout', None)
        cmd_out = cmd_out if cmd_out else getattr(
            proc, 'stderr', None)

        match = self._get_version(cmd_out.decode('utf8'))
        if not match:
            raise ValueError(
                'Unable to extract version via regexp "%s" from "%s"' % (
                    self.config.vre, cmd_out))
        self._compare_version(match.group())
        return True



"""Module for checking disk status.
"""

import logging
import shutil

from ox_mon.common import exceptions, configs, interface


class DiskUseTooHigh(exceptions.OxMonAlarm):
    """Exception for when disk usage is too high.
    """

    def __init__(self, loc: str, cur_pct_use: float, max_pct_use: float,
                 *args, **kwargs):
        msg = "Disk usage for %s too high: %.2f%% > %.2f%%" % (
            loc, cur_pct_use, max_pct_use)
        super().__init__(msg, *args, **kwargs)


class SimpleDiskChecker(interface.Checker):
    """Checker for testing disk usage.
    """

    @classmethod
    def options(cls):
        logging.debug('Making options for %s', cls)
        result = configs.BASIC_OPTIONS + [
            configs.OxMonOption(
                'tool', default='shutil', help=(
                    'Tool to use to determine disk usage. Default is '
                    'shutil which is sometimes flaky. If you install '
                    'psutil you can provide psutil for better results.')),
            configs.OxMonOption(
                'target', default='/', help=(
                    'Target location to check for usage (e.g., /).')),
            configs.OxMonOption(
                'max-used-pct', default=85.0, type=float, help=(
                    'Maximum percentage allowed for disk usage.'))
            ]
        return result

    def _get_du(self):
        "Get disk usage based on desired tool."

        if self.config.tool == 'shutil':
            result = shutil.disk_usage(self.config.target)
        elif self.config.tool == 'psutil':
            import psutil
            result = psutil.disk_usage(self.config.target)
        else:
            raise ValueError('Invalid tool "%s"' % self.config.tool)
        return result

    def _check(self):
        """Check disk usage.
        """

        disk_usage = self._get_du()
        current_percent_usage = (
            disk_usage.used/disk_usage.total) * 100

        if current_percent_usage > self.config.max_used_pct:
            raise DiskUseTooHigh(
                self.config.target,
                current_percent_usage, self.config.max_used_pct)
